/* abac_xml.c, specifically for GENI */

/* in xml, rc in general is, 1 is good, 0 or -1 is bad */

#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include <libxml/tree.h>
#include <libxml/xmlmemory.h>
#include <libxml/parser.h>

#ifndef XMLSEC_NO_XSLT
#include <libxslt/xslt.h>
#include <libxslt/security.h>
#endif /* XMLSEC_NO_XSLT */

#include <xmlsec/xmlsec.h>
#include <xmlsec/xmltree.h>
#include <xmlsec/xmldsig.h>
#include <xmlsec/base64.h>
#include <xmlsec/templates.h>
#include <xmlsec/crypto.h>
#include <xmlsec/list.h>

#include <openssl/sha.h>

extern void* abac_get_sha_from_nake_pem(char *npem, char **sha1);
extern void abac_split(char *string, char *delim, char **ret, int *num);
static int debug=0;

/* from Ted's reader */

/* for accessing GENI privilege credentials */
#define GENI_signed_credential "signed-credential"
#define GENI_credential "credential"
#define GENI_type "type"
#define GENI_serial "serial"
#define GENI_owner_gid "owner_gid"
#define GENI_owner_urn "owner_urn"
#define GENI_target_gid "target_gid"
#define GENI_target_urn "target_urn"
#define GENI_uuid "uuid"
#define GENI_expires "expires"
#define GENI_privileges "privileges"
#define GENI_privilege "privilege"
#define GENI_name "name"
#define GENI_can_delegate "can_delegate"

/* for accessing GENI abac credentials. signed-credential, credential, type and
 * expires are present as well.  */
#define GENI_rt0 "rt0"
#define GENI_version "version"

/* maximum credential stringlen */
#define CREDLEN 1024
/* Maximum version len */
#define VERSIONLEN 256

/* Call perror on str and exit */
static void fatal(const char *str) {
    perror(str);
    exit(20);
}

/* Hex dump the contents of buf (length len) to stderr.  For debugging */
static void dump_it(xmlChar *buf, int len) {
    int i;

    for (i=0; i < len; i++ ) 
	fprintf(stderr, "%02x ", buf[i] & 0xff);
    fprintf(stderr, "\n");
}

/* Convert a SHA1 hash to text for printing or use in strings.  keyid is the
 * hash (SHA_DIGEST_LENGTH bytes long) and text is the output string (at least
 * 2 * SHA_DIGEST_LENGTH +1 bytes long).  The caller is responsible for
 * allocating and deallocating each of them. */
static void sha1_to_text(xmlChar *keyid, xmlChar *text) {
    int i = 0;

    for (i=0; i < SHA_DIGEST_LENGTH; i++) 
	snprintf((char *) text+2*i, 3, "%02x", keyid[i] & 0xff);
}

/*
 * Returns the content pointer of the XML_TEXT node that is the child of the
 * node passed in.  Node must point to an XML_ELEMENT node.  the return value
 * is still part of node's XML document, so treat it as read only. */
static xmlChar *get_element_content(xmlNodePtr node) {
    xmlNodePtr text = NULL;

    if ( node->type != XML_ELEMENT_NODE ) return NULL;
    if ( !( text = node->children) ) return NULL;
    if ( text->type != XML_TEXT_NODE ) return NULL;
    return text->content;
}

/*Find the XML element named field that is a at or below node in the XML
 * document and return a copy of the base64 decoded content of its content.
 * For example, find key daya in a signature and return it base64decoded.  buf
 * is allocated and its length is returned in len.  Any values for buf and len
 * are ignored and overwritten. */
static int get_base64_field(xmlNodePtr node, xmlChar *field, 
	xmlChar **buf, int *len) {
    xmlChar *txt = NULL;

    *buf = NULL;

    if ( !(node = xmlSecFindNode(node, field, xmlSecDSigNs)))
	goto fail;

    if ( !(txt = get_element_content(node))) 
	    goto fail;
    
    *len = strlen((char *) txt);
    if ( !(*buf = malloc(*len))) 
	goto fail;

    if ( (*len = xmlSecBase64Decode(txt, *buf, *len)) < 0 ) 
	goto fail;

    return 1;
fail:
    if ( *buf) free(*buf);
    *len = 0;
    return 0;
}
/* Construct an ASN.1 header for a field of type h that is len bytes long.
 * Mosttly this creates the proper length encoding under ASN.1's length
 * encoding rules. buf will contain the new header (and the caller is
 * responsible for making sure it is at least 6 bytes long, though fewer are
 * usually used.  The length of the constructed header is returned. This is
 * used in creating a key hash from key data.*/
static int make_asn1_header(char h, size_t len, xmlChar *buf) {
    if ( len > 0x00ffffff) {
	buf[0] = h;
	buf[1] = 0x84;
	buf[2] = (len >> 24) & 0xff;
	buf[3] = (len >> 16) & 0xff;
	buf[4] = (len >> 8) & 0xff;
	buf[5] = (len) & 0xff;
	return 6;
    } else if ( len > 0x0000ffff ) {
	buf[0] = h;
	buf[1] = 0x83;
	buf[2] = (len >> 16) & 0xff;
	buf[3] = (len >> 8) & 0xff;
	buf[4] = (len) & 0xff;
	return 5;
    } else if ( len > 0x000000ff ) {
	buf[0] = h;
	buf[1] = 0x82;
	buf[2] = (len >> 8) & 0xff;
	buf[3] = (len) & 0xff;
	return 4;
    } else if ( len > 0x80 ) {
	buf[0] = h;
	buf[1] = 0x81;
	buf[2] = (len) & 0xff;
	return 3;
    } else {
	buf[0] = h;
	buf[1] = (len) & 0xff;
	return 2;
    }
}

/* Find the RSA key parameters in the KeyInfo section of the XML document
 * pointed to by doc, construct the ASN.1 encoding of that key and SHA1 hash
 * it.    This gives the standard keyid of that key.  keyid will be the binary
 * encoding of that (the bits of the hash)  sha1_to_text will turn it to text.
 * keyid must be at least SHA_DIGEST_LENGTH bytes long, and the caller is
 * responsible for it. This routine returns 1 on success and 0 on failure. */
static int get_keyid_from_keyinfo(xmlDocPtr doc, xmlChar *keyid) {
    xmlNodePtr root = NULL; /* XML document root */
    xmlNodePtr node = NULL; /* Scratch XML node */

    xmlChar b0[20];	    /* Header for the sequence */
    xmlChar b1[20];	    /* Header for the modulus */
    xmlChar b2[20];	    /* Header for the exponent */
    int l0 = 0;		    /* Header length for the sequence */
    int l1 = 0;		    /* Header length for the modulus */
    int l2 = 0;		    /* Header length for the exponent */

    xmlChar *modBuf = NULL; /* Bytes of the modulus */
    xmlChar *expBuf = NULL; /* Bytes of the exponent */
    int modLen = 0;	    /* Length of the modulus */
    int expLen = 0;	    /* Length of the exponent */

    SHA_CTX sha;	    /* SHA1 hash context */

    int rv = 0;		    /* return value */

    if ( !SHA1_Init(&sha)) goto fail;

    if ( !doc || !(root = xmlDocGetRootElement(doc)) ) 
	goto fail;

    /* Find the KeyInfo section to be the root of later searches */
    if ( !(node = xmlSecFindNode(root, 
		    xmlSecNodeKeyInfo, xmlSecDSigNs)))
	goto fail;

    /* Get the binary for the modulus and exponent */
    if ( !get_base64_field(node, (xmlChar *) "Modulus", &modBuf, &modLen)) 
	goto fail;
    if ( !get_base64_field(node, (xmlChar *) "Exponent", &expBuf, &expLen)) 
	goto fail;

    /* Construct the headers for modulus and exponent.  Another weird fact
     * about ASN.1 is that all the integers are signed, so if either the
     * modulus or exponent has the high order bit of its first byte set, the
     * ASN.1 encoding has a 0 byte prepended.  This code appends the 0 byte to
     * the header, which results in the same hash. */
    if ( modBuf[0] & 0x80 ) {
	l1 = make_asn1_header(0x02, modLen +1, b1);
	b1[l1++] = '\0';
    } else {
	l1 = make_asn1_header(0x02, modLen, b1);
    }

    if ( expBuf[0] & 0x80 ) {
	l2 = make_asn1_header(0x02, expLen +1, b2);
	b2[l2++] = '\0';
    } else {
	l2 = make_asn1_header(0x02, expLen, b2);
    }

    /* Sequence header: have to do it after we know the lengths of the inner
     * headers. */
    l0 = make_asn1_header(0x30, modLen + expLen + l1 + l2, b0);
    /* Hash it all up in parts */
    SHA1_Update(&sha, b0, l0);
    SHA1_Update(&sha, b1, l1);
    SHA1_Update(&sha, modBuf, modLen);
    SHA1_Update(&sha, b2, l2);
    SHA1_Update(&sha, expBuf, expLen);
    SHA1_Final(keyid, &sha);
    rv = 1;
fail:

    if (modBuf) free(modBuf);
    if (expBuf) free(expBuf);
    return rv;
}

/* Check the signature of either kind of credential - it'll work for basically
 * any signed XML. Returns true if the signature checks and false otherwise.
 * Takes a pointer to the XML document to check.  Similar to the examples at
 * http://www.aleksey.com/xmlsec/api/xmlsec-examples.html */
static int check_signature(xmlDocPtr doc) {
    xmlNodePtr root = NULL;	    /* Root XML node */
    xmlNodePtr node = NULL;	    /* Scratch XML node */
    xmlSecKeyInfoCtxPtr keyCtx = NULL;/* Key info context.  Used to parse
					 KeyInfo */
    xmlSecKeysMngrPtr keyMan = NULL;/* Key manager - used because we have
				       certificated. */
    xmlSecKeyPtr key = NULL;	    /* The key extracted */
    xmlSecDSigCtxPtr dsigCtx = NULL;/* Signature context */
    int rv = 0;			    /* Return value */

    if ( doc && !(root = xmlDocGetRootElement(doc)) ) 
	goto fail;

    /* Find the KeyInfo section to pull the keys out. */
    if ( !(node = xmlSecFindNode(root, 
		    xmlSecNodeKeyInfo, xmlSecDSigNs)))
	goto fail;

    /* Create and initialize key, key manager, and context */
    if ( !(key = xmlSecKeyCreate() ) )
	    goto fail;
    if ( !(keyMan = xmlSecKeysMngrCreate()) ) 
	goto fail;

    if ( xmlSecCryptoAppDefaultKeysMngrInit(keyMan) < 0)
	goto fail;

    if ( !(keyCtx = xmlSecKeyInfoCtxCreate(keyMan)) ) 
	goto fail;

    /* Do not check certificate signatures */
    keyCtx->flags |= XMLSEC_KEYINFO_FLAGS_X509DATA_DONT_VERIFY_CERTS;

    /* Gather up the key data */
    if ( xmlSecKeyInfoNodeRead(node, key, keyCtx) < 0 ) 
	goto fail;

    /* Set up the signature context and attack the keys */
    if ( !(dsigCtx = xmlSecDSigCtxCreate(NULL)))
	goto fail;

    dsigCtx->signKey = key;

    /* find the Signature section */
    if ( !(node = xmlSecFindNode(root, 
		    xmlSecNodeSignature, xmlSecDSigNs)))
	goto fail;

    /* Check it */
    if ( (rv = xmlSecDSigCtxVerify(dsigCtx, node)) < 0 ) 
	goto fail;

    /* Strangely xmlSecDSigCtxVerify can return success even if the status is
     * bad.  Check the status in the context explicitly and override the result
     * above if necessary.*/
    if ( dsigCtx->status != xmlSecDSigStatusSucceeded) 
	goto fail;

    return 1;
fail:
    if ( key ) xmlSecKeyDestroy(key);
    if ( keyCtx ) xmlSecKeyInfoCtxDestroy(keyCtx);
    if ( dsigCtx) xmlSecDSigCtxDestroy(dsigCtx);
    return 0;
}

/* Extract a sha from PEM blob */
static void extract_owner_sha1(xmlChar *pem, xmlChar **sha1) {
    void *x509=abac_get_sha_from_nake_pem((char *) pem, (char **)sha1); 
    if(debug) fprintf(stderr,"owner sha1 is (%s)..\n", *sha1);
}
static void extract_target_sha1(xmlChar *pem, xmlChar **sha1) {
    void *x509=abac_get_sha_from_nake_pem((char *) pem,(char **)sha1); 
    if(debug) fprintf(stderr,"group sha1 is (%s)..\n", *sha1);
}

/* Parse the content of the expires field and compare it to the time passed in
 * n.  If expires is earlier, return false, else true.  If n is null, compare
 * it to the current time. */
static int check_GENI_expires(xmlChar *expires, struct tm *n) {
    struct tm tv;   /* Parsed expires field */
    time_t now;	    /* Now in seconds since the epoch */
    time_t exp;	    /* expires in seconds since the epoch */

    if (n) now = mktime(n);
    else time(&now);

    strptime((char *) expires, "%FT%TZ", &tv);
    exp = timegm(&tv);

    return difftime(exp, now) > 0.0;
}

/* Convert a parsed privilege in a GENI privilege string into one or more ABAC
 * creds.  Rules are as in http://groups.geni.net/geni/wiki/TIEDCredentials .
 * keyid is  the issuer's keyid, text_owner is the owners keyid (XXX:a
 * placeholder) text_target is the target's keyid (XXX: a placeholder), priv is
 * the privilege being assigned, and delegate is true if it can be delegated.
 * carray is the output array of character strings that currently has *nc
 * entries in it.  If nc has nothing in it, insert the "speaks_for" delegation
 * rules.  Then add the privilege specific rules. On failure ***carray and its
 * contents are freed and *nc is set to zero.*/
static void add_privilege_credential_string(xmlChar *text_keyid, xmlChar *text_owner, 
	xmlChar *text_target, xmlChar *priv, int delegate, xmlChar ***carray,
	int *nc) {
    xmlChar **rv = *carray;	/* Local pointer to the array of strings */
    xmlChar **newrv = NULL;	/* Used to realloc the array of strings */
    int ncred = *nc;		/* Local copy of the number of creds in rv */
    int newc = (delegate) ? 3 : 1;  /* Number of new creds to add */
    int base = ncred;		/* First new credential index.  This advances
				   as we add creds to rv. */
    int i = 0;			/* Scratch */

    /* If rv is empty, add the speaks_for rules */
    if (base == 0 ) newc += 2;

    /* Resize rv */
    if (!(newrv = realloc(rv, (base + newc) * sizeof(xmlChar *))))
	goto fail;

    for ( i = base; i < base +newc; i ++) { 
	newrv[i] = NULL;
    }

    /* So fail works */
    rv = newrv;
    ncred = base + newc;

    /* Add speaks_for rules  if needed */
    if ( base == 0 ) {
	if ( !(rv[base] = malloc(CREDLEN))) goto fail;
	snprintf((char *) rv[base], CREDLEN, 
		"%s.speaks_for_%s <- %s.speaks_for_%s",
		text_keyid, text_owner, text_owner, text_owner);
	base++;
	if ( !(rv[base] = malloc(CREDLEN))) goto fail;
	snprintf((char *) rv[base], CREDLEN, 
		"%s.speaks_for_%s <- %s",
		text_keyid, text_owner, text_owner);
	base++;
    }

    /* The assignemnt of priv.  Always happens */
    if ( !(rv[base] = malloc(CREDLEN))) goto fail;
    snprintf((char *) rv[base], CREDLEN, 
	    "%s.%s_%s <- %s.speaks_for_%s",
	    text_keyid, priv, text_target, text_keyid, text_owner);
    base++;
    /* Add delegation rules */
    if ( delegate ) {
	if ( !(rv[base] = malloc(CREDLEN))) goto fail;
	snprintf((char *) rv[base], CREDLEN, 
		"%s.%s_%s <- %s.can_delegate_%s_%s.%s_%s",
		text_keyid, priv, text_target, text_keyid, priv, 
		text_target, priv, text_target);
	base++;
	if ( !(rv[base] = malloc(CREDLEN))) goto fail;
	snprintf((char *) rv[base], CREDLEN, 
		"%s.can_delegate_%s_%s <- %s",
		text_keyid, priv, text_target, text_owner);
	base++;
    }
    /* And return new values */
    *carray = rv;
    *nc = ncred;
    return;
fail:
    if ( rv ) {
	/* Delete all the allocations, ours or not, and clear the caller's 
	 * variables */
	for (i = 0; i < ncred; i++) 
	    if (rv[i]) free(rv[i]);
	free(rv);
    }
    *carray = NULL;
    *nc = 0;
}

/* Parse a GENI privilege credential (that has already had its signature
 * checked) and return the RT0 strings that the credential is encoded as.  The
 * return value is an array of strings, zero-terminated (like argv) that holds
 * the RT0 strings.  It is NULL on failure. */
static xmlChar **parse_privilege(xmlDocPtr doc) {
    xmlNodePtr root = NULL;	/* XML root node */
    xmlNodePtr node = NULL;	/* XML scratch node */
    xmlNodePtr owner = NULL;	/* XML owner_gid node */
    xmlNodePtr expires = NULL;	/* XML expires node */
    xmlNodePtr target = NULL;	/* XML target_gid node */
    xmlNodePtr privs = NULL;	/* XML privileges node */
    xmlNodePtr priv = NULL;	/* XML privilege node - used to iterate */
    xmlChar keyid[SHA_DIGEST_LENGTH];	/* Issuer key SHA1 */
    xmlChar text_keyid[2*SHA_DIGEST_LENGTH+1];/* Issuer keyid as text */
    xmlChar *owner_sha1 = NULL;         /* owner gid as text */
    xmlChar *target_sha1 = NULL;        /* target gid as text */
    xmlChar **newrv = NULL;	/* Used to realloc rv to add the NULL
				   terminator*/
    xmlChar **rv = NULL;	/* return value */
    int ncred = 0;		/* number of creds in rv, incase we need to
				   deallocate it */
    int i = 0;			/* scratch */

    if ( doc && !(root = xmlDocGetRootElement(doc)) ) 
	goto fail;

    /* Get the issuer keyid */
    if ( !get_keyid_from_keyinfo(doc, keyid)) 
	goto fail;
    sha1_to_text(keyid, text_keyid);

    /* Find the various fields of interest */
    if ( !(node = xmlSecFindNode(root, (xmlChar *) GENI_credential, NULL)))
	goto fail;

    /* Make sure this is not expired */
    if ( !(expires = xmlSecFindNode(node, (xmlChar *) GENI_expires, NULL)))
	goto fail;

    if ( !check_GENI_expires(get_element_content(expires), NULL))
	goto fail;

    /* XXX: owner and target will be X.509 pem files from which we need to
     * extract keyids for add_privilege_credential_string.  I didn't replciate
     * that code which lives in libabac. Code to get those keyids needs to be
     * added. */
    if ( !(owner = xmlSecFindNode(node, (xmlChar *) GENI_owner_gid, NULL)))
	goto fail;
    extract_owner_sha1(get_element_content(owner),&owner_sha1);

    if ( !(target = xmlSecFindNode(node, (xmlChar *) GENI_target_gid, NULL)))
	goto fail;
    extract_target_sha1(get_element_content(target),&target_sha1);

    if ( !(privs = xmlSecFindNode(node, (xmlChar *) GENI_privileges, NULL)))
	goto fail;

    /* Iterate through the privileges, parsing out names and can_delegate and
     * generating the strings from it. */
    for (priv = privs->children; priv; priv = priv->next) {
	/* reinitialized every time around */
	xmlNodePtr n = NULL;
	xmlChar *name = NULL;
	int delegate = -1;

	/* Privilege node */
	if ( priv->type != XML_ELEMENT_NODE || 
		strcmp((char *) priv->name, (char *) GENI_privilege) ) {
	    fprintf(stderr, "Bad node!\n");
	    goto fail;
	}
	/* looking for name and can_delegate */
	for (n = priv->children; n; n = n->next) {
	    if ( n->type != XML_ELEMENT_NODE ) 
		continue;
	    if ( !strcmp((char *) n->name, (char *) GENI_name)) {
		name = get_element_content(n);
		continue;
	    }
	    if ( !strcmp((char *) n->name, (char *) GENI_can_delegate)) {
		xmlChar *boolean = get_element_content(n);

		if ( !strcmp((char *) boolean, "true") ||
			!strcmp((char *) boolean, "1") ) {
		    delegate = 1;
		} else if ( !strcmp((char *) boolean, "false") ||
			!strcmp((char *) boolean, "0") ) {
		    delegate = 0;
		} else {
		    fprintf(stderr, "Unknown delegation value %s", boolean);
		}
	    }
	}
	/* Found both name and can_delegate, add the RT0 to rv and ncred */
	if ( name && delegate != -1 ) {
	    add_privilege_credential_string(text_keyid, 
		    (xmlChar *) owner_sha1, (xmlChar *) target_sha1, name,
		    delegate, &rv, &ncred);
	    if ( !rv ) goto fail;
	}
    }

    /* Add the terminating NULL */
    if (!(newrv = realloc(rv, sizeof(xmlChar*)*(ncred+1))))
	goto fail;

    newrv[ncred] = NULL;
    return newrv;

fail:
    /* Throw away all of rv if there's an error */
    if ( rv ) {
	for (i = 0; i < ncred; i++) 
	    if (rv[i]) free(rv[i]);
	free(rv);
    }
    return NULL;
}

/* Compare the version string (major.minor) to the major and minor numbers, and
 * return true if the version string is >= the major minor pair. */
static int check_GENI_abac_version(xmlChar *version, int major, int minor) {
    xmlChar lversion[VERSIONLEN];   /* Buffer to parse version in */
    xmlChar *lp = &lversion[0];	    /* strsep needs a pointer it can modify */
    int maj = 0;		    /* Parsed major number */
    int min = 0;		    /* Parsed minor number */
    char *values[2];
    int ret=0;

    snprintf((char *) lp, VERSIONLEN, "%s", (char *) version);
    abac_split(lp, ".", values, &ret);
    if(ret==2) {
        /* split successful, values[0] is the major number,
           values[1] is the minor, now we parse integers*/
        if ( *values[0] == '\0') return 0;
        maj = strtol(values[0], NULL, 10);
        if ( *values[1] == '\0') return 0;                            
        min = strtol(values[1], NULL, 10);    
    } else return 0;                                                  
                                                                      
    if (maj > major ||              
           (maj == major && min > minor) ||                           
               (maj == major && min == minor) ) return 1;             
    else return 0;                                                    
}  
/*
 * Parse an ABAC credential (that has had its signature checked.  Make sure it
 * has not expired and that its version is at least 1.0 (XXX: parameterize
 * that), and return the RT0 it encodes as the only entry in an NULL-terminated
 * array of strings (just like parse_privilege). On failure return NULL. */
xmlChar **parse_abac(xmlDocPtr doc) {
    xmlNodePtr root = NULL;	/* XML root node */
    xmlNodePtr node = NULL;	/* XML credential node */
    xmlNodePtr rt0 = NULL;	/* XML rt0 node */
    xmlNodePtr expires = NULL;	/* XML expires node */
    xmlNodePtr version = NULL;	/* XML version node */
    xmlChar keyid[SHA_DIGEST_LENGTH];/* issuer SHA1 hash */
    xmlChar text_keyid[2*SHA_DIGEST_LENGTH+1];/* Issuer keyid in text */
    xmlChar *txt;		/* rt0 content */
    xmlChar **rv = NULL;	/* return value */
    int ncred = 0;		/* number of creds in rv */
    int i = 0;			/* Scratch (used only in fail:)  */

    if ( doc && !(root = xmlDocGetRootElement(doc)) ) 
	goto fail;

    /* Get the issuer keyid */
    if ( !get_keyid_from_keyinfo(doc, keyid)) 
	goto fail;
    sha1_to_text(keyid, text_keyid);
    /* get the various nodes of interest */
    if ( !(node = xmlSecFindNode(root, (xmlChar *) GENI_credential, NULL)))
	goto fail;
    if ( !(expires = xmlSecFindNode(node, (xmlChar *) GENI_expires, NULL)))
	goto fail;
    if ( !(rt0 = xmlSecFindNode(node, (xmlChar *) GENI_rt0, NULL)))
	goto fail;
    if ( !(version = xmlSecFindNode(node, (xmlChar *) GENI_version, NULL)))
	goto fail;

    /* Check the version and expiration */
    if ( !check_GENI_abac_version(get_element_content(version), 1, 0)) 
	goto fail;
    if ( !check_GENI_expires(get_element_content(expires), NULL))
	goto fail;

    /* read the RT0 and return it */
    if ( !(txt = get_element_content(rt0)) ) goto fail;

    if ( !(rv = malloc(2 * sizeof(xmlChar *)))) goto fail;
    if (!(rv[0] = malloc(strlen((char *) txt)+1))) goto fail;
    ncred++;

    strcpy((char *) rv[0], (char *) txt);
    rv[1] = NULL;

    return rv;
fail:
    if ( rv ) {
	for (i = 0; i < ncred; i++) 
	    if (rv[i]) free(rv[i]);
	free(rv);
    }
    return NULL;
}

/* Check and parse a GENI credential.  Return the new RT0 rules in a
 * NULL-terminated array of strings.  If the signature or parsing fails, return
 * NULL. Demultiplexed to parse_privilege or parse_abac to do the parsing and
 * uses check_signature to confirm the sig. */
static xmlChar **read_credential(char *infile, xmlChar **xml) {
    xmlDocPtr doc = xmlParseFile(infile);   /* Parse the document */
    xmlNodePtr node = NULL;		    /* XML scratch node */
    xmlChar *text = NULL;		    /* Text of the type field */
    xmlChar **rv = NULL;		    /* return value from parse_* */

    if ( !check_signature(doc) ) 
	goto fail;
    /* Parse out the type field */
    if ( !(node = xmlDocGetRootElement(doc)) ) 
	goto fail;
    if ( !(node = xmlSecFindNode(node, (xmlChar *) GENI_credential, NULL)))
	goto fail;
    if ( !(node = xmlSecFindNode(node, (xmlChar *) GENI_type, NULL)))
	goto fail;

    if ( !(text = get_element_content(node)) ) 
	goto fail;

    /* Demux on type */
    if ( !strcmp((char *) text, "privilege")) {
	rv = parse_privilege(doc);
    } else if ( !strcmp((char *) text, "abac")) {
	rv = parse_abac(doc);
    } else { 
	goto fail;
    }
    int len=0;
    xmlDocDumpMemoryEnc(doc, xml, &len, "UTF-8");
    if(len == 0)
       goto fail;

fail:
    xmlFreeDoc(doc);
    return rv;
}


/* from Ted's writer */

/* maximum rt0 stringlen */
#define CREDLEN 1024
/* Maximum version len */
#define VERSIONLEN 256

/* XML template for a new RT0 credential.  Fill in the RT0 to store (XML
 * escaped) and the expiration time (formated using strftime_format and
 * strftime). */
xmlChar *template = (xmlChar *) (
"<signed-credential>\n"
"    <credential xml:id=\"ref0\">\n"
"	<type>abac</type>\n"
"	<version>1.0</version>\n"
"	<expires>%s</expires>\n"
"	<rt0>%s</rt0>\n"
"    </credential>\n"
"    <signatures>\n"
"	<Signature xmlns=\"http://www.w3.org/2000/09/xmldsig#\">\n"
"	    <SignedInfo>\n"
"		<CanonicalizationMethod Algorithm=\"http://www.w3.org/2001/10/xml-exc-c14n#\"/>\n"
"		<SignatureMethod Algorithm=\"http://www.w3.org/2000/09/xmldsig#rsa-sha1\"/>\n"
"		<Reference URI=\"#ref0\">\n"
"		    <Transforms>\n"
"			<Transform Algorithm=\"http://www.w3.org/2000/09/xmldsig#enveloped-signature\"/>\n"
"		    </Transforms>\n"
"		    <DigestMethod Algorithm=\"http://www.w3.org/2000/09/xmldsig#sha1\"/>\n"
"		    <DigestValue/>\n"
"		</Reference>\n"
"	    </SignedInfo>\n"
"	    <SignatureValue/>\n"
"	    <KeyInfo>\n"
"	      <KeyValue/>\n"
"		<X509Data>\n"
"		    <X509Certificate/>\n"
"		    <X509SubjectName/>\n"
"		    <X509IssuerSerial/>\n"
"		</X509Data>\n"
"	    </KeyInfo>\n"
"	</Signature>\n"
"    </signatures>\n"
"</signed-credential>");

char *strftime_fmt = "%FT%TZ";
#define EXPIRESLEN 20

/* Return a copy of str with > < and & replaced by &gt; &lt; and &amp; for
 * embedding in XML.  Caller is responsible for deallocating the return value
 * using xmlFree().
 */
static xmlChar *minimal_xml_escaping(xmlChar *str) {
    /* A quickie translation table with the character to escape, the output
     * string and the length of the output in it. The table is initialized with
     * the three translations we want. */
    static struct esc {
	xmlChar c;
	xmlChar *esc;
	int l;
    } escapes[] = {
	{ (xmlChar) '<', (xmlChar *) "&lt;", 4 }, 
	{ (xmlChar) '>', (xmlChar *) "&gt;", 4},
	{ (xmlChar) '&', (xmlChar *) "&amp;", 5},
	{ (xmlChar) '\0', NULL, 0 },
    };

    xmlChar *rv = NULL;	    /* Return value */
    xmlChar *p = NULL;	    /* Scratch (walking str) */
    xmlChar *q = NULL;	    /* Scratch (walking rv) */
    struct esc *e = NULL;   /* Scratch for walking escapes */
    int len = 0;	    /* Length of rv */

    /* Walk str and calculate how long the escaped version is */
    for ( p = str; *p ; p++) {
	int foundit = 0;
	for ( e = escapes; !foundit && e->c ; e++ ) {
	    if ( *p == e->c ) {
		len += e->l;
		foundit = 1;
	    }
	}
	if ( !foundit ) len++;
    }
    /* Allocate the new string */
    q = rv = (xmlChar *) xmlMalloc(len+1);
    /* copy str to rv, escaping when necessary */
    for ( p = str; *p ; p++) {
	int foundit = 0;
	for ( e = escapes; !foundit && e->c ; e++ ) {
	    if ( *p == e->c ) {
		strncpy((char *) q, (char *) e->esc, e->l);
		q += e->l;
		foundit = 1;
	    }
	}
	if ( !foundit ) *q++ = *p;
    }
    /* terminate rv */
    *q = '\0';
    return rv;
}

/* Make an new GENI abac credential with the rt0 rule that expires secs from
 * now.  cert is the PEM encoded X.509 of the issuer's certificate as a string.
 * certlen is the length of cert.  Returns the XML. Caller is responsible for
 * freeing it. */
static char *make_credential(xmlChar *rt0, int secs, xmlSecByte *cert, 
	xmlSecSize certlen) {
    xmlDocPtr doc = NULL;	/* parsed XML document */
    xmlNodePtr root = NULL;	/* Root of the document */
    xmlNodePtr signNode = NULL;	/* <Signature> element */
    xmlSecDSigCtxPtr dsigCtx = NULL;  /* Signing context */
    xmlChar *rt0_escaped = minimal_xml_escaping(rt0); /* RT0, escaped */
    xmlChar *rv = NULL;		/* return value */
    time_t exp;			/* expriation time (seconds since epoch) */
    struct tm exp_tm;		/* expiration for formatting */
    char estr[EXPIRESLEN+1];	/* String with formatted expiration */
    char *temp = NULL;		/* populated XML template */
    int len = 0;		/* length of the populated template (temp) */

    if ( !rt0_escaped) goto fail;

    /* Calculate the length of the populated template and allocate it */
    len = strlen((char *) template)+EXPIRESLEN+strlen((char *) rt0_escaped)+1;

    if ( !(temp = malloc(len)) )
	goto fail;

    /* format expiration */
    time(&exp);
    exp += secs;
    gmtime_r(&exp, &exp_tm);

    if (strftime(estr, EXPIRESLEN+1, strftime_fmt, &exp_tm) == 0 ) 
	goto fail;

    /* Populate template with  expiration and escaped rt0 */
    snprintf(temp, len, (char *) template, estr, (char *) rt0_escaped);

    /* parse the populated template */
    if ( !(doc = xmlParseMemory(temp, len))) 
	goto fail;

    if (!(root = xmlDocGetRootElement(doc)) )
	goto fail;

    /* Find the signature element for the Signing call */
    signNode = xmlSecFindNode(root, xmlSecNodeSignature, xmlSecDSigNs);

    /* Create the context */
    if ( !(dsigCtx = xmlSecDSigCtxCreate(NULL))) 
	goto fail;

    /* Assign the key (a PEM key) */
    if (!(dsigCtx->signKey = xmlSecCryptoAppKeyLoadMemory(cert, certlen,
		    xmlSecKeyDataFormatPem, NULL, NULL, NULL)) )
	goto fail;

    /* Load the certificate */
    if ( xmlSecCryptoAppKeyCertLoadMemory(dsigCtx->signKey, cert, certlen,
		xmlSecKeyDataFormatPem) < 0)
	goto fail;

    /* Sign it */
    if ( xmlSecDSigCtxSign(dsigCtx, signNode) < 0)
	goto fail;

    /* Store the signed credential to rv */
    xmlDocDumpMemoryEnc(doc, &rv, &len, "UTF-8");
fail:
    /* clean up */
    if (dsigCtx) 
	xmlSecDSigCtxDestroy(dsigCtx);
    if ( doc) xmlFreeDoc(doc);
    if (rt0_escaped) xmlFree(rt0_escaped);
    if ( temp) free(temp);
    return (char *) rv;
}

#define CERTSIZE 4096
/* Read a file into a the dynamically allocated array buf.  Whatever was in buf
 * is discarded w/o freeing so it should initially be NULL and len 0. On return
 * buf contains as much of the contents of filename as possible and len is the
 * length of it.  The caller must free() buf. buf is zero terminated.*/
static void read_cert(char *filename, char **buf, int *len) {
    FILE *cert = NULL;
    char *nbuf = NULL;
    int cap = CERTSIZE;
    int r = 0;

    *len = 0; 
    if (!(*buf = malloc(CERTSIZE))) return;

    if ( !(cert = fopen(filename, "r"))) 
	goto fail;

    /* Read cert, expanding buf as necessary */
    while ( (r = fread(*buf + *len, 1, cap - *len, cert)) != 0) {
	*len += r;
	if ( *len == cap ) {
	    if ( !(nbuf = realloc(*buf, cap + CERTSIZE)) ) 
		goto fail;
	    memcpy(nbuf, *buf, cap);
	    cap += CERTSIZE;
	    *buf = nbuf;
	}
    }
    /* Extend if we read exactly cap bytes so we can terminate the string */
    if ( *len == cap ) {
	if ( !(nbuf = realloc(*buf, cap + CERTSIZE)) ) 
	    goto fail;
	memcpy(nbuf, *buf, cap);
	cap += CERTSIZE;
	*buf = nbuf;
    }
    (*buf)[*len] = '\0';
    return;
fail:
    if ( *buf) free(*buf);
    *buf = NULL;
    *len = 0;
    return;
}


/******** helper functions used by libabac **********************/

/* Function to disable libXML2's debugging */
static void _nullGenericErrorFunc(void* ctxt, const char* msg, ...) { return; }

/* Straight off http://www.aleksey.com/xmlsec/api/xmlsec-examples.html .
 * Voodoo.  But call it. */
int init_xmlsec() {
    /* Init xmlsec library */
    if(xmlSecInit() < 0) {
        fprintf(stderr, "Error: xmlsec initialization failed.\n");
        return(-1);
    }

    /* Check loaded library version */
    if(xmlSecCheckVersion() != 1) {
        fprintf(stderr,
		"Error: loaded xmlsec library version is not compatible.\n");
        return(-1);
    }

    /* Load default crypto engine if we are supporting dynamic
     * loading for xmlsec-crypto libraries. Use the crypto library
     * name ("openssl", "nss", etc.) to load corresponding 
     * xmlsec-crypto library.
     */
#ifdef XMLSEC_CRYPTO_DYNAMIC_LOADING
    if(xmlSecCryptoDLLoadLibrary(BAD_CAST XMLSEC_CRYPTO) < 0) {
        fprintf(stderr, "Error: unable to load default xmlsec-crypto library. "
		"Make sure\n" 
		"that you have it installed and check shared libraries path\n"
		"(LD_LIBRARY_PATH) envornment variable.\n");
        return(-1);     
    }
#endif /* XMLSEC_CRYPTO_DYNAMIC_LOADING */

    /* Init crypto library */
    if(xmlSecCryptoAppInit(NULL) < 0) {
        fprintf(stderr, "Error: crypto initialization failed.\n");
        return(-1);
    }

    /* Init xmlsec-crypto library */
    if(xmlSecCryptoInit() < 0) {
        fprintf(stderr, "Error: xmlsec-crypto initialization failed.\n");
        return(-1);
    }
    /* Turn off the built in debugging */
    xmlThrDefSetGenericErrorFunc(NULL, _nullGenericErrorFunc);
    xmlSetGenericErrorFunc(NULL, _nullGenericErrorFunc);

    return 0;
}

int deinit_xmlsec() {
  /* no op for now */
    return 0;
}


/* to cast the parameters */
char **my_read_credential(char *infile, char **xml) {
    return (char **)read_credential(infile, (xmlChar **)xml);
}

char *my_make_credential(char* rt0, int secs, char* cert, int certlen)
{
    return make_credential((xmlChar *)rt0, secs, (xmlSecByte *)cert, (xmlSecSize)certlen);
}

/* parse the xml blob and extract keyid */ 
char *get_keyid_from_xml(char *xml) {
    xmlDocPtr doc=xmlParseMemory(xml,strlen(xml));
    xmlChar keyid[SHA_DIGEST_LENGTH];   /* Issuer key SHA1 */
    xmlChar text_keyid[2*SHA_DIGEST_LENGTH+1];/* Issuer keyid as text */
    char *ret=NULL;

    /* Get the issuer keyid */
    if ( !get_keyid_from_keyinfo(doc, keyid))
        goto fail;
    sha1_to_text(keyid, text_keyid);
    ret=strdup((char *)text_keyid);
fail:
    xmlFreeDoc(doc);
    return ret;
}

/* parse xml and get the expected expiration time and returns 
 * (expiration time-current time)
 */
long get_validity_from_xml(char *xml) {
    xmlDocPtr doc=xmlParseMemory(xml,strlen(xml));
    xmlNodePtr node = NULL;                 /* XML scratch node */
    xmlNodePtr expires = NULL;  /* XML expires node */
    struct tm tv;   /* Parsed expires field */
    time_t now;     /* Now in seconds since the epoch */
    time_t exp;     /* expires in seconds since the epoch */
    long dtime=0;

    if ( !(node = xmlDocGetRootElement(doc)) )
        goto fail;

    if ( !(expires = xmlSecFindNode(node, (xmlChar *) GENI_expires, NULL)))
        goto fail;

    xmlChar *etime=get_element_content(expires);
    time(&now);

    strptime((char *) etime, "%FT%TZ", &tv);
    exp = timegm(&tv);
    dtime=difftime(exp, now);

fail:
    xmlFreeDoc(doc);
    return dtime;
}

/* parse xml structure and extract the attribute rules */
char **get_rt0_from_xml(char *xml) {
    xmlDocPtr doc=xmlParseMemory(xml,strlen(xml));
    xmlNodePtr node = NULL;                 /* XML scratch node */
    xmlChar *text = NULL;                   /* Text of the type field */
    xmlChar **rv = NULL;                    /* return value from parse_* */
    
    if ( !check_signature(doc) )
        goto fail;
    /* Parse out the type field */
    if ( !(node = xmlDocGetRootElement(doc)) )
        goto fail;
    if ( !(node = xmlSecFindNode(node, (xmlChar *) GENI_credential, NULL)))
        goto fail;
    if ( !(node = xmlSecFindNode(node, (xmlChar *) GENI_type, NULL)))
        goto fail;

    if ( !(text = get_element_content(node)) )
        goto fail;

    /* Demux on type */
    if ( !strcmp((char *) text, "privilege")) {
        rv = parse_privilege(doc);
    } else if ( !strcmp((char *) text, "abac")) {
        rv = parse_abac(doc);
    } else {
        goto fail;
    }
fail:
    xmlFreeDoc(doc);
    return (char **)rv;
}


